#include "hash.h"

#include <stdio.h>

#include <string.h>

#include <stdlib.h>

#include <iostream>
#include <fstream>

#include <time.h>

#include "debug.h"

#include "bitboard.h"

#include "move.h"

#include "movelist.h"

#include "board.h"

#include "utils.h"

#include "input.h"

#include "tree.h"

#include "learn.h"

#include "movelist.h"

#include "history.h"

#pragma pack(1)

using namespace std;

class OBEntry {
public:
	HASHCODE hashcode;	// upper 16 bits parent's 16 bits upper hashcode
						// lower 48 bits children's 48 bits lower hashcode
	unsigned short int white_wins;
	unsigned short int black_wins;
	unsigned short int draws;
	int score;			// position score after a n ply search
};

#pragma pack()


const int INVALID_SCORE = 666666;


const int OB_SIZE = 1 << 16;
const int OB_MASK = OB_SIZE - 1;

void ScoreBook(int wtm, int ply);

inline int ComputeIndex(HASHCODE& code){
	//return (int)((((code>>48)<<2)|(code & 3)) & OB_MASK);
	return (int)((code>>48) & OB_MASK);
}

class OBElement {
public:
	OBEntry *data;	// dynamic array of OpeningBook entries with the same parent
	int count;
	int size;		// number of entries allocated
	int delta;		// amount to increase the array when 

	OBElement(){
		count = 0;
		size = 5;
		data = new OBEntry[size];
		Assert(data);
		memset(data, 0, sizeof(OBEntry)*size);
		delta = 5;
	}

	~OBElement(){
    	delete [] data;
    }

	void Grow(int amount);

	void Add(OBEntry* entry);
	
	OBEntry* New(HASHCODE hashcode);

	OBEntry* Find(HASHCODE& hashcode);
	int Exists(HASHCODE& hashcode);
};

void OBElement::Add(OBEntry* entry){
	OBEntry* ob = Find(entry->hashcode);
	int ww = entry->white_wins; ww += ob->white_wins;
	int bw = entry->black_wins; bw += ob->black_wins;
	int dr = entry->draws; dr += ob->draws;

	ob->white_wins	= (ww < 65535) ? ww : 65535;
	ob->black_wins	= (bw < 65535) ? bw : 65535;
	ob->draws		= (dr < 65535) ? dr : 65535;

	ob->score = INVALID_SCORE;
}


	void OBElement::Grow(int amount){
		Assert(amount > 0);
		OBEntry* new_data = new OBEntry[size+amount];
		Assert(new_data);
		memset(new_data, 0, sizeof(OBEntry)*(size+amount));
		memcpy(new_data, data, sizeof(OBEntry)*size);
		delete[] data;
		data = new_data;
		size += amount;
	}

	OBEntry* OBElement::New(HASHCODE hashcode){
		if (count == size){
			Grow(delta);
			//delta *= 2; // grows by 2x next time
		} 

		Assert(count < size);
	
		return &(data[count++]);
	}

	OBEntry* OBElement::Find(HASHCODE& hashcode){
		#if 0

		for (int i = count-1; i >= 0; --i){		
			if (data[i].hashcode == hashcode)
				return data+i;
		}
		OBEntry* e = New(hashcode);
		e->hashcode = hashcode;
		return e;
		
		#else
		if (count > 0){
			int left,right;

			left = 0;
			right = count-1;

			while (right - left > 1){				
				int middle = (left+right)>>1;

				if (data[middle].hashcode <= hashcode)
					left = middle;
				else
					right = middle;
			}	

			if (data[left].hashcode == hashcode)
				return data+left;
			if (data[right].hashcode == hashcode)
				return data+right;


			if (data[count-1].hashcode < hashcode){
				OBEntry* e = New(hashcode);
				e->hashcode = hashcode;
				return e;
			}

			while (left > 0 && data[left].hashcode > hashcode)
				left--;
			while (left < count && data[left].hashcode < hashcode)
				left++;
			
			New(hashcode);

			for (int i = count-1; i > left; i--)
				data[i] = data[i-1];
			data[left].hashcode = hashcode;
			data[left].white_wins = data[left].black_wins = data[left].draws = 0;

			/*
			for (i = 0; i+1 < count; i++){
				Assert(data[i].hashcode < data[i+1].hashcode);
			}
			*/

			return data+left;
			
		} else {
			OBEntry* e = New(hashcode);
			e->hashcode = hashcode;
			return e;
		}
		#endif
	}


	int OBElement::Exists(HASHCODE& hashcode){
		for (int i = count-1; i >= 0; --i){		
			if (data[i].hashcode == hashcode){
				// kludge : a position exists if it has not been processed
				return (data[i].score == INVALID_SCORE);
			}
		}
		return 0;
	}


class OBIndexElement {
public:
	int start;
	int size;
};

class OB {
protected:
	OBElement list[OB_SIZE];
	OBIndexElement index[OB_SIZE];

	FILE* fp;

public:
	
	OB();
	~OB();

	int MaxCount();

	int Count();

	int Probe(int wtm);
	OBEntry* ReadCluster(HASHCODE code, int first_time = 0);

	void Add(HASHCODE code, char result);

	void Add(OBElement& entries);

	void Write(char* data, char* index);
	void Read(char* data, char* index);

	int Exists(HASHCODE code);

	void SetScore(HASHCODE code, int score);
};


OB ob;



void OB::SetScore(HASHCODE code, int score){
	int idx = ComputeIndex(code);
	OBEntry* oe = list[idx].Find(code);
	oe->score = score;
}

int OB::Count(){
	int x = 0;
	for (int i = 1; i < OB_SIZE; i++)
		x += list[i].count;
	return x;
}


int OB::MaxCount(){
	int max = list[0].count;
	for (int i = 1; i < OB_SIZE; i++)
		if (list[i].count > max)
			max = list[i].count;
	return max;
}

void OB::Add(OBElement& entries){
	for (int i = 0; i < entries.count; i++){	
		int idx = ComputeIndex(entries.data[i].hashcode);
		OBElement& e = list[idx];
		e.Add(entries.data+i);
	}
}

int OB::Exists(HASHCODE code){
	int idx = ComputeIndex(code);
	return list[idx].Exists(code);
}



void OB::Add(HASHCODE code, char result){
	int idx = ComputeIndex(code);
	OBEntry* e = list[idx].Find(code);

	if (result == 1){
		if (e->white_wins < 65535)
			e->white_wins++;
	} else if (result == -1){
		if (e->black_wins < 65535)
			e->black_wins++;
	}else {
		if (e->draws < 65535)
			e->draws++;
	}
}

void PrintBB(BITBOARD b, ofstream& cout, int level = 0){
	if (b == 0) {
		if (level == 0) 
			cout << '0';
		return;
	}
	PrintBB(b / 10, cout, level+1);
	cout << (unsigned int)(b % 10);
}


void OB::Write(char* data, char* index_name){
	FILE* fp;
	int i;

	int max_cluster_size = 0;

	ofstream cout;

	cout.open("stats.txt");

	cout << "INDEX :" << endl;

	// construct index
	int count = 0;

	int position = 0;
	for (i = 0; i < OB_SIZE; i++){

		count += list[i].count;
		
		index[i].start = position;
		index[i].size = list[i].count * sizeof(OBEntry);

		max_cluster_size = Max(max_cluster_size, list[i].count);

		cout <<	"	" << i << " start = " << index[i].start << " size = " << list[i].count << endl;;
		position += index[i].size;
	}


	cout << count << " unique positions" << endl;
	cout << max_cluster_size << " max cluster size" << endl;

	fp = fopen(index_name, "wb");
	if (!fp){
		cout << "Cannot open index for opening book " << index_name << endl;
		return;
	}

	fwrite(index, sizeof(index), 1, fp);
		
	fclose(fp);

	fp = fopen(data, "wb");
	if (!fp){
		cout << "Cannot open opening book " << data << " for writing" << endl;
		return;
	}
	for (i = 0; i < OB_SIZE; i++){
		fwrite(list[i].data, sizeof(OBEntry),list[i].count, fp);
		for (int x = 0; x < list[i].count; x++){
			Assert(ComputeIndex(list[i].data[x].hashcode) == i);
		}		
	}

	fclose(fp);

	cout << endl << endl;


	#if 0
	cout << "Table dump :" << endl;

	count = 0;

	for (i = 0; i < OB_SIZE; i++){

		for (int j = 0; j < list[i].count; j++){
			cout << count++ << "	";
			cout << "Hashcode : ";
			PrintBB(list[i].data[j].hashcode, cout);
			cout << "	wins : " << list[i].data[j].white_wins << " " << list[i].data[j].black_wins << " " << list[i].data[j].draws;
			cout << "	score : " << list[i].data[j].score;
			cout << endl;
		}
	}
	#endif

	cout.close();
}

void OB::Read(char* data, char* index_file){
	fp = fopen(index_file, "rb");
	if (!fp){
		cout << "Cannot open index file" << endl;
		return;
	}

	int count = fread(this->index, sizeof(OBIndexElement), OB_SIZE, fp);
	if (count != OB_SIZE){
		fp = NULL;
		cout << "Index file size isn't right" << endl;
		return;
	}


	fclose(fp);

	fp = fopen(data, "rb");
	if (!fp){
		cout << "Cannot open index book " << endl;
		return;
	}	




	#if 0
	{	
		ofstream cout;
		cout.open("stats2.txt");

		FILE* fp;

		OBEntry oe[256];

		memset(oe, 666, sizeof(oe));

		fp = fopen(data, "rb");

		int count = 0;

		cout << sizeof(OBEntry) << endl;

		while (!feof(fp)){
			int n = fread(oe, sizeof(OBEntry), 256, fp);

			cout << n << "items read" << endl;
			
			for (int i = 0; i < n; i++){
				cout << count++ << "	";
				cout << "Hashcode : ";
				PrintBB(oe[i].hashcode, cout);
				cout << "	wins : " << oe[i].white_wins << " " << oe[i].black_wins << " " << oe[i].draws;
				cout << "	score : " << oe[i].score;
				cout << endl;
			}
		}

		cout.close();

		fclose(fp);
	}
	#endif
}



const int MAX_BUF_SIZE = 5000;
struct {
	HASHCODE hashcode;
	char result;
} temp_buf[MAX_BUF_SIZE];


//static OBEntry temp_ob[MAX_BUF_SIZE];
static OBElement temp_ob;
//static int temp_ob_n = 0;

void ResetTempOB(){
	temp_ob.count = 0;
}

void AddDataToTempOB(int count){
	for (int x = 0; x < count; x++){
		HASHCODE code = temp_buf[x].hashcode;

		//Print(code); cout << endl;

		int result = temp_buf[x].result;

		OBEntry* ptr = temp_ob.Find(code);
		
		if (result == 1){
			if (ptr->white_wins < 65535)
				ptr->white_wins++;
		} else if (result == -1){
			if (ptr->black_wins < 65535)
				ptr->black_wins++;
		}else {
			if (ptr->draws < 65535)
				ptr->draws++;
		}
	}
}


void AddTempOBToOB(){
	ob.Add(temp_ob);
}


int ob_count = 0;

void BuildBook(char* temp_name, char* book_name, char* index_file){

	int cnt = 0;

	FILE* fp = fopen(temp_name, "rb");
	if (!fp){
		cout << "Cannot open temp file " << temp_name << endl;
		return;
	}
	
	temp_ob.Grow(MAX_BUF_SIZE);
	
	while (1) {
		int count = fread(temp_buf, sizeof(temp_buf[0]), MAX_BUF_SIZE, fp);
		/*for (int i = 0; i < count; i++)
			ob.Add(temp_buf[i].hashcode, temp_buf[i].result);
		*/

		ResetTempOB();
		AddDataToTempOB(count);
		AddTempOBToOB();

		cnt += count;

		cout << cnt << " positions so far" << " ( max count = " << ob.MaxCount() << ")" << endl;

		if (count < MAX_BUF_SIZE)
			break;
	}

	ob_count = ob.Count();
	cout << ob_count << " unique positions" << endl;


	#if 0
	board.Reset();
	ClearHT();
	
	ScoreBook(WHITE, 0);
	#endif

	ob.Write(book_name, index_file);

}


OB::OB(){
	fp = NULL;
}

OB::~OB(){
	if (fp)
		fclose(fp);
}


OBEntry* OB::ReadCluster(HASHCODE code, int first_time){
	static OBEntry cluster[180];
	static int size;

	int idx = ComputeIndex(code);

	if (first_time){
		Assert(fp);
		fseek(fp, index[idx].start, SEEK_SET);

		size = index[idx].size / sizeof(OBEntry);
	
		if (size > 180){
			cout << "Too big opening book cluster" << endl;
			return NULL;
		}

		if (fread(cluster, sizeof(OBEntry), size, fp) != (unsigned)size){
			cout << "there was some error reading an opening book cluster" << endl;
			return NULL;
		}
	}

	if (!first_time)
	for (int i = 0; i < size; i++){
		Assert(ComputeIndex(cluster[i].hashcode) == idx);
		/*
		cout << i << " = ";
		Print(cluster[i].hashcode);
		cout << " " << (int)((cluster[i].hashcode >> 48) &OB_MASK) << endl;
		*/
		if (cluster[i].hashcode == code)
			return cluster + i;
	}

	//Assert(false);

	return NULL;
}


static float Percent(int total, int fraction){
	return (total > 0) ? ((fraction * 100.0f)/(float)total) : 0;
}

static void sort_list_reverse(int* moves, int* score, int n){
	int i;
	int done;
	int tmp;

	do {
		done = 1;
		for (i = 0; i+1 < n; i++)
			if (score[i] < score[i+1]){
				done = 0;
				tmp = score[i]; score[i] = score[i+1]; score[i+1] = tmp;
				tmp = moves[i]; moves[i] = moves[i+1]; moves[i+1] = tmp;
			}
	} while(!done);
}

inline int MoveScore(OBEntry* e, int wtm){
	int result = wtm ? e->white_wins : e->black_wins;
	return result;
}


int OB::Probe(int wtm){
	int i;
	int* m;
	int score;

	if (!fp) {
		//cout << "Book Probe : No book loaded !" << endl;
		return 0;
	}

	static int m_list[1024];
	static int real_score[1024];
	memset(m_list, 0, sizeof(m_list));

	static int move_score[512];

	HASHCODE parent = board.hashcode;
	HASHCODE child;
	HASHCODE hashcode;

	int* last = board.GenerateAllValidMoves(m_list, wtm);
	
	ReadCluster(parent, 1);
	
	SaveBoardInfo(0);

	
	// Step 1 : score moves based on master book
	int* moves_so_far, n_moves_so_far;
	GetMoveList(&moves_so_far, n_moves_so_far);

	memset(move_score, 0, sizeof(move_score));
	
	if (!input_flags.edited_position)
		ScoreMoves(m_list, move_score, last-m_list,moves_so_far, n_moves_so_far);

	// Step 2 : add to the previous score the score of the moves in the book
	for (m = m_list; m < last; m++){
		//cout << "Probing move : "; PrintMove(*m); cout << endl;

		board.Make(*m, 0, wtm);

		/*
		if (From(*m) == D2 && To(*m) == D4 && Piece(*m) == PAWN){
			board.Print();

			cout << "Bingo !" << endl;

			cout << "parent hashcode = "; Print(parent); cout << endl;
			cout << "child  hashcode = "; Print(board.hashcode); cout << endl;

			HASHCODE hc = Shl(Shr(parent,48),48) + Shr(Shl(board.hashcode,16),16);

			cout << "result hashcode = "; Print(hc); cout << endl;
		}
		*/


		child = board.hashcode;
		hashcode = Shl(Shr(parent,48),48) + Shr(Shl(child,16),16);
		OBEntry* e = ReadCluster(hashcode);
		if (e) {
			score = MoveScore(e, wtm);
			real_score[m-m_list] = e->score;

			#if 0
				cout << "Move : "; PrintMove(*m);
				cout << "score = " << score;
				cout << endl;
			#endif

			if (score > 1) 
				move_score[m-m_list] += score;
		} else {
			real_score[m-m_list] = 0;
			//cout << "Not found !"  << endl;
		}

		board.Unmake(*m, 0, wtm);
	}


	// Step 3 : sort the list and keep from it only the moves scoring close to the first move
	int n = last-m_list;

	sort_list_reverse(m_list, move_score, n);

	if (n == 0)	return 0;
	int counter = 0;
	for (i = 0; i < n; i++)
		if (move_score[i] > 0)
			counter++;

	if (counter == 0) return 0;

	// print the move list
	if (input_flags.analyze_mode){
		cout << "1.	0	0	0	0 Book moves : ";
	} else
		cout << "Best book moves in this position : ";
	for (m = m_list; m < last; m++)	{
		if (move_score[m-m_list] != 0){
			int sc = real_score[m-m_list];
			int dec = (sc < 0) ? ((-sc) % 1000) : (sc % 1000);
			PrintMove(*m); cout << "(" << move_score[m-m_list];
			//cout << " " << (sc/1000) << "." << (dec/10);
			cout << ") ";
		}
	}
	cout << endl;

	const int MARGIN = 100;
	
	for (i = 1; i < n; i++){
		if (move_score[i]+MARGIN <= move_score[0] || move_score[i] <= 3){
			n = i;
			break;
		}
	}


	// if in analyze mode return dummy move
	if (input_flags.analyze_mode && n)
		return m_list[0];
	
	if (n == 0) return 0;
	
	srand((unsigned)time(0));	
	int move = m_list[(Random32()+rand()) % n];

	//cout << "will return : " ;PrintMove(move); cout << endl;
	return move;
}


void ReadBook(char* book_name, char* index_name){
	ob.Read(book_name, index_name);

	LearnBook("learn.txt", "learn.bin");
}


int ProbeBook(int wtm){
	return ob.Probe(wtm);
}


static int m_list[256];

extern int current_move[];
extern int current_ply;

void DumpBook(int wtm, int ply){
	HASHCODE parent = board.hashcode;
	HASHCODE child;
	HASHCODE hashcode;

	int move_list[256];

	int* last = board.GenerateCaptures(move_list, wtm);
	last = board.GenerateNonCaptures(last, wtm);
	*last = 0;
	
	parent = board.hashcode;
	ob.ReadCluster(parent, 1);

	int written = 0;
	int found = 0;

	SaveBoardInfo(ply);
	
	for (int* m = move_list; m < last; m++){
		m_list[ply] = *m;

		current_ply = ply;
		board.Make(*m, ply, wtm);

		/*if (!Check(wtm))*/{
			child = board.hashcode;
			hashcode = (Shl(Shr(parent,48),48)) + (Shr(Shl(child,16),16));
			ob.ReadCluster(parent, 1);
			OBEntry* e = ob.ReadCluster(hashcode);

			if (e){
					DumpBook(1-wtm, ply+1);
					found = 1;
				/*	
				for (int i = 0; i < ply; i++) cout << " ";
				PrintMove(*m);
				cout<< endl;
				if (ply < 2)
					DumpBook(1-wtm, ply+1);
					*/
			} 
		}

		current_ply = ply;
		board.Unmake(*m, ply, wtm);
	}

	if (ply > 0 && !written && !found){
		for (int i = 0; i < ply; i++)
			PrintMove(m_list[i]);
		cout << endl;
		written = 1;
	}

}


const int MAX_PLY_FOR_SCORING = 3;


int score_count = 0;

void ScoreBook(int wtm, int ply){
	HASHCODE parent = board.hashcode;
	HASHCODE child;
	HASHCODE hashcode;

	int move_list[256];

	int* last = board.GenerateCaptures(move_list, wtm);
	last = board.GenerateNonCaptures(last, wtm);
	*last = 0;
	
	parent = board.hashcode;

	int written = 0;
	int found = 0;

	SaveBoardInfo(ply);

	for (int* m = move_list; m < last; m++){
		m_list[ply] = *m;

		current_ply = ply;
		board.Make(*m, ply, wtm);

		if (!Check(wtm)){
			child = board.hashcode;
			hashcode = (Shl(Shr(parent,48),48)) + (Shr(Shl(child,16),16));

			if (ob.Exists(hashcode)){

					cout << "Computing the score for the following position ";
					cout << "(" << ++score_count << " out of " << ob_count << ")" << endl;
					//board.Print();

					/*
					for (int i = 0; i < 8; i++){
						for (int j = 0; j < 8; j++)
							cout << board.sq[i*8+j];

						cout << endl;
					} */

					input_flags.time_left = 2560000; //plenty
					
					BackupTree(ply+1);

					ResetHistory();
					ClearPawnHT();

					int dummy;
					board.Analyse(1-wtm, dummy, MAX_PLY_FOR_SCORING);

					RestoreTree(ply+1);	

					ob.SetScore(hashcode, input_flags.search_score);

					board.age = (board.age + 1) & 7;

					ScoreBook(1-wtm, ply+1);

					board.age = (board.age + 8 - 1) & 7;

					found = 1;
			} 
		}

		current_ply = ply;
		board.Unmake(*m, ply, wtm);
	}

	/*
	if (ply > 0 && !written && !found){
		for (int i = 0; i < ply; i++)
			PrintMove(m_list[i]);
		cout << endl;
		written = 1;
	}
	*/

}

